{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE OverloadedRecordDot #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE StandaloneDeriving #-}
{-# OPTIONS_GHC -fno-warn-orphans #-}

module Jon.Server (JonAPI, jonSwaggerDoc, server) where

import Control.Lens
import Control.Monad.IO.Class (liftIO)
import Data.Aeson (FromJSON, ToJSON(toJSON))
import Data.Text (Text)
import Data.Int (Int32)
import Data.Swagger
import Database.PostgreSQL.Simple (Connection)
import GHC.Generics (Generic)
import Servant
import Servant.Swagger (toSwagger)

import Jon.Garfield.Types

import qualified Jon.Garfield.Queries as Queries

-- API and types

type JonAPI =
    "rpc" :>
        (    "getUnsoundBarcodes"   :> ReqBody '[JSON] GetUnsoundBarcodesP
                                    :> Post '[JSON] [UnsoundBarcodeDTO]
        :<|> "getOverviewItems"     :> ReqBody '[JSON] GetOverviewItemsP
                                    :> Post '[JSON] [OverviewItemDTO]
        :<|> "getActiveItems"       :> ReqBody '[JSON] GetActiveItemsP
                                    :> Post '[JSON] [OverviewItemDTO]
        :<|> "getLocations"         :> Post '[JSON] [Location]
        )

data GetUnsoundBarcodesP = GetUnsoundBarcodesP
    { location :: Int32
    } deriving (Generic, FromJSON, ToSchema)

data UnsoundBarcodeDTO = UnsoundBarcodeDTO
    { barcode   :: Text
    , name      :: Text
    , entries   :: Int
    , unitsLeft :: Int
    } deriving (Generic, ToJSON, ToSchema)

data GetOverviewItemsP = GetOverviewItemsP
    { location :: Int32
    } deriving (Generic, FromJSON, ToSchema)

data OverviewItemDTO = OverviewItemDTO
    { overview :: Overview
    , item     :: InventoryItem
    } deriving (Generic, ToJSON, ToSchema)

mkOverviewItemDTO = uncurry OverviewItemDTO

data GetActiveItemsP = GetActiveItemsP
    { barcode  :: Text
    , location :: Int32
    } deriving (Generic, FromJSON, ToSchema)

-- Orphan instances for database types

instance ToJSON InventoryItemId where
    toJSON = toJSON . (.unInventoryItemId)
instance ToSchema InventoryItemId where
    declareNamedSchema _ = declareNamedSchema (Proxy :: Proxy Int32)

instance ToJSON InventoryItemGroupId where
    toJSON = toJSON . (.unInventoryItemGroupId)
instance ToSchema InventoryItemGroupId where
    declareNamedSchema _ = declareNamedSchema (Proxy :: Proxy Int32)

instance ToJSON TaxGroupId where
    toJSON = toJSON . (.unTaxGroupId)
instance ToSchema TaxGroupId where
    declareNamedSchema _ = declareNamedSchema (Proxy :: Proxy Int32)

instance ToJSON LocationId where
    toJSON = toJSON . (.unLocationId)
instance ToSchema LocationId where
    declareNamedSchema _ = declareNamedSchema (Proxy :: Proxy Int32)

deriving instance ToJSON Overview
deriving instance ToSchema Overview

deriving instance ToJSON InventoryItem
deriving instance ToSchema InventoryItem

deriving instance ToJSON Location
deriving instance ToSchema Location

-- server

server :: Connection -> Server JonAPI
server conn = getUnsoundBarcodes :<|> getOverviewItems :<|> getActiveItems :<|> getLocations
    where
        getUnsoundBarcodes :: GetUnsoundBarcodesP -> Handler [UnsoundBarcodeDTO]
        getUnsoundBarcodes params = do
            rows <- liftIO $ Queries.runSelect conn $ Queries.unsoundBarcodes $ mkLocationId params.location
            pure $ map mkUnsoundBarcodeDTO rows
            where
                mkUnsoundBarcodeDTO (a, b, c, d) = UnsoundBarcodeDTO a b (fromIntegral c) (fromIntegral d)

        getOverviewItems :: GetOverviewItemsP -> Handler [OverviewItemDTO]
        getOverviewItems params = do
            rows <- liftIO $ Queries.runSelect conn $ Queries.overviewItemsByLocation $ mkLocationId params.location
            pure $ map mkOverviewItemDTO rows
        
        getActiveItems :: GetActiveItemsP -> Handler [OverviewItemDTO]
        getActiveItems params = do
            rows <- liftIO $ Queries.runSelect conn $ Queries.activeItems params.barcode $ mkLocationId params.location
            pure $ map mkOverviewItemDTO rows

        getLocations :: Handler [Location]
        getLocations = do
            liftIO $ Queries.runSelect conn Queries.locations

jonSwaggerDoc = toSwagger (Proxy :: Proxy JonAPI)
    & info . title .~ "jon API"
